#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <sys/ioctl.h>
//#include <machine/param.h>

//#include <machine/db_machdep.h>
#define ELFSIZE DB_ELFSIZE

#include <linux/elf.h>
#include <asm/user.h>
typedef int vaddr_t;
#define Elf_Ehdr       Elf32_Ehdr
#define Elf_Phdr       Elf32_Phdr
#define Elf_Shdr       Elf32_Shdr
#define mips_round_page(x)	((((unsigned)(x)) + NBPG - 1) & ~(NBPG-1))
#define mips_trunc_page(x)	((unsigned)(x) & ~(NBPG-1))


//#include <sys/exec_elf.h>
//#include <machine/bootinfo.h>
//#include "memio.h"
#include <linux/akmemio.h>

int main __P((int, char *[]));
int seek_kernel __P((int, caddr_t*, caddr_t*, caddr_t*, int));
void usage __P((void));

int verbose;

int
main(int argc, char **argv)
{

	int kernel, akmem;
	int ch, reboot, unload;
	char akname[] = "/dev/akmem";
	struct akmem_alloc alloc_info;
	struct akmem_exec exec_info;
	caddr_t start, end, entry;

	reboot = 0;
	unload = 0;
	verbose = 0;
	while ((ch = getopt(argc, argv, "ruv")) != -1)
		switch(ch) {
		case 'r':
			reboot = 1;
			break;
		case 'u':
			unload = 1;
			break;
		case 'v':
			verbose = 1;
			break;
		default:
			usage();
		}
	argc -= optind;
	argv += optind;

	if (!unload) {
		if (argc == 0) {
			usage();
			exit(1);
		}
		if ((kernel = open(argv[0], O_RDONLY)) < 0) {
			(void)fprintf(stderr, "can't open %s\n", argv[0]);
			exit(1);
		}
		if (seek_kernel(kernel, &start, &end, &entry, 0)) {
			(void)fprintf(stderr, "can't seek kernel\n");
			close(kernel);
			exit(1);
		}
		exec_info.entry = entry;
	}

	if ((akmem = open(akname, O_RDWR)) < 0) {
		(void)fprintf(stderr, "can't open %s\n", akname);
		exit(1);
	}

	if (unload) {
		if (close(akmem) < 0) {
			(void)fprintf(stderr, "can't close %s\n", akname);
			exit(1);
		}
		exit(0);
	}

	alloc_info.nsegments = 1;
	alloc_info.segments = (struct akmem_segment*)
	    malloc(alloc_info.nsegments * sizeof(struct akmem_segment));
	alloc_info.segments[0].start = (void *)mips_trunc_page(start);
	alloc_info.segments[0].end = (void *)mips_round_page(end);
	if (verbose) {
		(void)printf("allocate: %x-%x(%d pages)\n",
		    (u_int)alloc_info.segments[0].start,
		    (u_int)alloc_info.segments[0].end, (u_int)(end - start) / NBPG);
	}

	if (ioctl(akmem, AKMEMIOCTL_ALLOC, &alloc_info) < 0) {
		(void)fprintf(stderr, "can't allocate physical memory\n");
		close(akmem);
		close(kernel);
		exit(1);
	}
	if (verbose) {
		(void)printf("loading %s...", argv[0]);
		fflush(stdout);
	}
	if (seek_kernel(kernel, &start, &end, &entry, akmem) != 0) {
		(void)fprintf(stderr, "can't load kernel\n");
		close(kernel);
		close(akmem);
		exit(1);
	}
	close(kernel);
	if (verbose)
		(void)printf("done\n");
	exec_info.argc = argc;
	exec_info.argv = argv;
	//exec_info.env = NULL;
	if (ioctl(akmem, AKMEMIOCTL_COMMIT, &exec_info) < 0) {
		(void)fprintf(stderr, "can't commit another kmem\n");
		close(akmem);
		exit(1);
	}
	close(akmem);
	if (reboot) {
		execl("/sbin/reboot", "/sbin/reboot", (char *)0);
		/* not reachable */
		exit(1);
	}
	exit(0);
}

int
seek_kernel(fd, start, end, entry, load)
	int fd, load;
	caddr_t *start, *end, *entry;
{
	Elf_Ehdr elfx, *elf = &elfx;
	int i, first;
	Elf_Shdr *shtbl = NULL;
	Elf_Phdr *phtbl = NULL;
	caddr_t min_addr, max_addr;
	char page_buf[NBPG];

	if (lseek(fd, 0, SEEK_SET) == -1)  {
		(void)fprintf(stderr,"seek error\n");
		goto error_cleanup;
	}
	if (read(fd, (void*)elf, sizeof(Elf_Ehdr)) != sizeof(Elf_Ehdr)) {
		(void)fprintf(stderr,"read header error\n");
		goto error_cleanup;
	}

	if ((phtbl = (Elf_Phdr *)malloc(sizeof(*phtbl) * elf->e_phnum)) == NULL
	|| (shtbl = (Elf_Shdr *)malloc(sizeof(*shtbl) * elf->e_shnum)) == NULL) {
		(void)fprintf(stderr,"malloc() error\n");
		goto error_cleanup;
	}

	if (lseek(fd, elf->e_phoff, SEEK_SET) == -1)  {
		(void)fprintf(stderr,"seek for program header table error\n");
		goto error_cleanup;
	}
	if (read(fd, (void *)phtbl, sizeof(Elf_Phdr) * elf->e_phnum)
			!= (int)(sizeof(Elf_Phdr) * elf->e_phnum)) {
		(void)fprintf(stderr,"read program header table error\n");
		goto error_cleanup;
	}

	if (lseek(fd, elf->e_shoff, SEEK_SET) == -1)  {
		(void)fprintf(stderr,"seek for segment header table error.\n");
		goto error_cleanup;
	}
	if (read(fd, (void *)shtbl, sizeof(Elf_Shdr) * elf->e_shnum)
			!= (int)(sizeof(Elf_Shdr) * elf->e_shnum)) {
		(void)fprintf(stderr,"read segment header table error\n");
		goto error_cleanup;
	}

	first = 1;
	for (i = 0; i < elf->e_phnum; i++) {
		if (phtbl[i].p_type != PT_LOAD) {
			continue;
		}

		if (first || max_addr
			< (caddr_t)(phtbl[i].p_vaddr + phtbl[i].p_memsz)) {
			max_addr
			= (caddr_t)(phtbl[i].p_vaddr + phtbl[i].p_memsz);
		}
		if (first || (caddr_t)phtbl[i].p_vaddr < min_addr) {
			min_addr = (caddr_t)phtbl[i].p_vaddr;
		}

		if (load) {
			vaddr_t file_end, mem_end, offset;
			int load_len, fill_len;
			size_t len;

			if (lseek(fd, phtbl[i].p_offset, SEEK_SET) == -1)  {
				goto error_cleanup;
			}

			offset = phtbl[i].p_vaddr;
			if (lseek(load, offset, SEEK_SET) == -1)  {
				goto error_cleanup;
			}
			file_end = offset + phtbl[i].p_filesz;
			mem_end = offset + phtbl[i].p_memsz;
			while (offset < mem_end) {
				vaddr_t next_offset;

				next_offset = mips_round_page(offset + 1);
				next_offset = next_offset < mem_end ? next_offset : mem_end;
				len = next_offset - offset;
				load_len = file_end < offset ? 0 : file_end - offset;
				if (load_len < len)
					fill_len = len - load_len;
				else {
					load_len = len;
					fill_len = 0;
				}
				if (read(fd, page_buf, load_len) != load_len)  {
					(void)fprintf(stderr, "can't load kernel to buf\n");
					goto error_cleanup;
				}
				memset(page_buf + load_len, 0, fill_len);
				if (write(load, page_buf, len) != len) {
					(void)fprintf(stderr, "can't load kernel\n");
					goto error_cleanup;
				}
				offset = next_offset;
			}
		}
		first = 0;
	}
	if (first) {
		goto error_cleanup;
	}

	if (phtbl) free(phtbl);
	if (shtbl) free(shtbl);

	if (start) *start = min_addr;
	if (end) *end = max_addr;
	if (entry) *entry = (caddr_t)elf->e_entry;
	return (0);

error_cleanup:
	if (phtbl) free(phtbl);
        if (shtbl) free(shtbl);
	return (1);
}

void
usage()
{
	extern char *__progname;

	(void)fprintf(stderr, "usage: %s [-u | [-rv] <kernel> [<boot options>]]\n",
	    __progname);
	exit(1);
}
